import base64 as _base64


def encode(text: str, encoding: str = "utf-8") -> str:
	return _base64.b64encode(text.encode(encoding=encoding)).decode(encoding=encoding)


def decode(text: str, encoding: str = "utf-8") -> str:
	return _base64.b64decode(text.encode(encoding=encoding)).decode(encoding=encoding)



from __future__ import annotations

import re
from abc import ABC, abstractmethod
from collections.abc import Callable
from datetime import datetime, timedelta, timezone
from datetime import tzinfo as _TzInfo
from typing import Self


class Segment(ABC):
	priority: int = 0
	"""## Higher value `=` applied earlier (NOT try_parsed earlier, try_parsing is dependant on the order of Parser.segment_types: tuple[type[Segment]])"""

	@classmethod
	@abstractmethod
	def try_parse(cls, text: str, /) -> Self | None:
		"""Attempt to parse the given text as this segment type. Return an instance if successful, or None otherwise."""

	@abstractmethod
	def apply(self, dt: datetime, now: datetime, segments: list[Segment], /) -> datetime:
		"""Apply this segment to the datetime dt and return the new datetime."""


class DateSegment(Segment):
	priority = 20

	DATE_REGEX = re.compile(r"^(?P<day>\d{1,2})-" r"(?P<month>\d{1,2})?" r"(?:-(?P<year>\d{1,4})?)?$")

	def __init__(self, day: int, month: int | None, year: int | None):
		self.day = day
		self.month = month
		self.year = year

	@classmethod
	def try_parse(cls, text: str, /) -> Self | None:
		m = cls.DATE_REGEX.fullmatch(text)
		if not m:
			return None

		day = int(m.group("day"))
		month_str = m.group("month")
		year_str = m.group("year")
		month = int(month_str) if month_str and month_str != "" else None
		year = int(year_str) if year_str and year_str != "" else None
		return cls(day, month, year)

	def apply(self, dt: datetime, now: datetime, segments: list[Segment], /) -> datetime:
		try:
			additional_years = 0
			additional_months = 0

			tries = 10

			while 1:
				if additional_months > 12:
					additional_months -= 12
					additional_years += 1

				try:
					dt = dt.replace(
						year=(self.year if self.year is not None else dt.year) + additional_years,
						month=(self.month if self.month is not None else dt.month) + additional_months,
						day=self.day,
					)

					break
				except ValueError as e:
					if str(e) != "day is out of range for month" and self.day <= 31:
						raise

					tries -= 1
					if tries <= 0:
						raise

					if self.month is not None and self.year is not None:
						raise

					if self.month is None:
						additional_months += 1
						continue

					additional_years += 1
					continue

			if self.month is not None and self.year is not None:
				return dt

			if self.month is not None:
				while dt <= now:
					dt = dt.replace(year=dt.year + 1)

				return dt

			while dt <= now:
				new_month = dt.month + 1
				new_year = dt.year

				if new_month > 12:
					new_month -= 12
					new_year += 1

				dt = dt.replace(month=new_month, year=new_year)

			return dt  # noqa: TRY300

		except ValueError as e:
			raise ValueError(f"Invalid date while applying date segment: {self}") from e

	def __repr__(self):
		return f"DateSegment(day={self.day}, month={self.month}, year={self.year})"


class TimeSegment(Segment):
	priority = 10

	TIME_REGEX = re.compile(r"^(?P<hour>\d{1,2}):" r"(?P<minute>\d{1,2})?" r"(?:\:(?P<second>\d{1,2})?)?$")

	def __init__(self, hour: int, minute: int, second: int):
		self.hour = hour
		self.minute = minute
		self.second = second

	@classmethod
	def try_parse(cls, text: str, /) -> Self | None:
		m = cls.TIME_REGEX.fullmatch(text)
		if not m:
			return None
		hour_str = m.group("hour")
		minute_str = m.group("minute")
		second_str = m.group("second")
		hour = int(hour_str) if hour_str != "" else 0
		minute = int(minute_str) if (minute_str is not None and minute_str != "") else 0
		second = int(second_str) if (second_str is not None and second_str != "") else 0

		if not (0 <= hour <= 23):
			raise ValueError("Hour must be in 0-23")
		if not (0 <= minute <= 59):
			raise ValueError("Minute must be in 0-59")
		if not (0 <= second <= 59):
			raise ValueError("Second must be in 0-59")
		return cls(hour, minute, second)

	def apply(self, dt: datetime, now: datetime, segments: list[Segment], /) -> datetime:
		before = dt

		try:
			dt = dt.replace(hour=self.hour, minute=self.minute, second=self.second, microsecond=0)
		except ValueError as e:
			raise ValueError("Invalid time after applying time segment") from e

		if not any(isinstance(segment, DateSegment | WeekdaySegment) for segment in segments):
			if dt <= before:
				dt = dt + timedelta(days=1)

		return dt

	def __repr__(self):
		return f"TimeSegment(hour={self.hour}, minute={self.minute}, second={self.second})"


class WeekdaySegment(Segment):
	priority = 15

	# fmt: off
	WEEKDAYS = [
		'mo',           'tu',      'we',        'th',       'fr',     'sa',       'su',
		'mon',          'tue',     'wed',       'thu',      'fri',    'sat',      'sun',
		'monday',       'tuesday', 'wednesday', 'thursday', 'friday', 'saturday', 'sunday',

		"po",           "wt",      "sr",        "cz",       "pi",     "so",       "ni",
		"pn",           "wt",      "śr",        "cz",       "pt",     "so",       "nd",
		"pon",          "wto",     "śro",       "czw",      "pią",    "sob",      "nie",
		"pon",          "wto",     "sro",       "czw",      "pia",    "sob",      "nie",
		"poniedzialek", "wtorek",  "sroda",     "czwartek", "piatek", "sobota",   "niedziela",
		"poniedziałek", "wtorek",  "środa",     "czwartek", "piątek", "sobota",   "niedziela",
	]
	# fmt: on

	def __init__(self, weekday: int):
		self.weekday = weekday

	@classmethod
	def try_parse(cls, text: str, /) -> Self | None:
		key = text.lower()
		if key in cls.WEEKDAYS:
			return cls(cls.WEEKDAYS.index(key) % 7)
		return None

	def apply(self, dt: datetime, now: datetime, segments: list[Segment], /) -> datetime:
		current_weekday = dt.weekday()
		days_ahead = (self.weekday - current_weekday) % 7
		if days_ahead == 0:
			days_ahead = 7  # always jump forward at least one week if already on that day
		return dt + timedelta(days=days_ahead)

	def __repr__(self):
		return f"WeekdaySegment(weekday={self.weekday})"


class RobostrSegment(Segment):
	PAIR_REGEX = re.compile(r"([-+]?(?:\d+(?:\.\d*)?|\.\d+))([a-zA-Z]+)")

	# fmt: off
	UNITS = {
		's':       (t_second := 1),
		'sec':     t_second,
		'secs':    t_second,
		'sex':     t_second, # >:3 # I am so funny, i know
		'second':  t_second,
		'seconds': t_second,

		'm':       (t_minute := 60*t_second),
		'min':     t_minute,
		'mins':    t_minute,
		'minute':  t_minute,
		'minutes': t_minute,

		'h':     (t_hour := 60*t_minute),
		'hr':    t_hour,
		'hrs':   t_hour,
		'hour':  t_hour,
		'hours': t_hour,

		'd':    (t_day := 24*t_hour),
		'day':  t_day,
		'days': t_day,

		'w':     (t_week := 7*t_day),
		'week':  t_week,
		'weeks': t_week,

		'y':     (t_year := 365*t_day), # assuming the 'y' unit always means the non-leap year...
		'year':  t_year,
		'years': t_year,

		'pul':   (t_pull := (11*t_hour + 30*t_minute)),
		'pull':  t_pull,
		'puls':  t_pull,
		'pulls': t_pull,
		'card':  t_pull,
		'cards': t_pull,

		'res':     (t_rescue := 6*t_hour),
		'reses':   t_rescue,
		'resees':  t_rescue,
		'rescue':  t_rescue,
		'rescues': t_rescue,

		'decade':    (t_decade := 10 * t_year),
		'decades':    t_decade,
	}
	# fmt: on

	def __init__(self, seconds: float):
		self.seconds = seconds

	@classmethod
	def try_parse(cls, text: str, /) -> Self | None:
		total = 0.0
		pos = 0
		for m in cls.PAIR_REGEX.finditer(text):
			if m.start() != pos:
				return None
			num_str = m.group(1)
			unit_str = m.group(2).lower()
			try:
				value = float(num_str)
			except ValueError as e:
				raise ValueError(f"Invalid number in robostr segment: {num_str}") from e
			if unit_str not in cls.UNITS:
				raise ValueError(f"Unknown unit '{unit_str}' in robostr segment")
			total += value * cls.UNITS[unit_str]
			pos = m.end()
		if pos != len(text) or pos == 0:
			return None
		return cls(total)

	def apply(self, dt: datetime, now: datetime, segments: list[Segment], /) -> datetime:
		return dt + timedelta(seconds=self.seconds)

	def __repr__(self):
		return f"RobostrSegment(seconds={self.seconds})"


class TimestrParser:
	segment_types: tuple[type[Segment]]
	respect_segment_priority: bool
	priorty_overrides: dict[type[Segment], int]

	def __init__(self, *segment_types: type[Segment], respect_segment_priority: bool = True, priority_overrides: dict[type[Segment], int] = None):
		self.segment_types = segment_types
		self.respect_segment_priority = respect_segment_priority
		self.priorty_overrides = priority_overrides if priority_overrides is not None else {}

	def parse(self, input_str: str, tz: timezone = datetime.now().astimezone().tzinfo, *, _datetime_now: Callable[[_TzInfo], datetime] = datetime.now) -> datetime:
		"""Given an input string (e.g. "1h20m!13:") and a timezone, return a datetime representing the datetime after now(tz=tz) + the amount of deltatime stored as segments in the input_str.

		The input string is processed as follows:
			1. Remove all spaces and split by '!'
			2. try_parse each segment (weekday, date, time, or robostr by default)
				- In the order they were passed to the parser's __init__
			3. Starting from now (with timezone tz), apply the segments by their priority from highest to lowest, by default it's:
				- date segments (if any; missing parts are inherited from now),
				- weekday segments (jumping forward to the next occurrence, even if landed spot-on),
				- time segments (missing pieces default to 0, eg. 13: -> 13:00:00, 13:45 -> 13:45:00),
				- robostr segments (adding a relative offset).
		"""
		cleaned = input_str.replace(" ", "")
		parts = [part for part in cleaned.split("!") if part != ""]

		all_segments: list[Segment] = []

		for part in parts:
			seg = None
			for seg_class in self.segment_types:
				candidate = seg_class.try_parse(part)
				if candidate is not None:
					seg = candidate
					break
			if seg is None:
				raise ValueError(f"Unrecognized segment: {part}")

			all_segments.append(seg)

		root_dt = now_dt = _datetime_now(tz).replace(microsecond=0)

		if self.respect_segment_priority:
			all_segments_priority_adjusted = sorted(all_segments, key=lambda seg: self.priorty_overrides.get(type(seg), seg.priority), reverse=True)
		else:
			all_segments_priority_adjusted = all_segments

		for seg in all_segments_priority_adjusted:
			root_dt = seg.apply(root_dt, now_dt, all_segments)

		return root_dt


SEGMENTS = (WeekdaySegment, DateSegment, TimeSegment, RobostrSegment)

timestr_priority_adjusted = TimestrParser(*SEGMENTS)
timestr = TimestrParser(*SEGMENTS, respect_segment_priority=False)



from typing import TYPE_CHECKING, Literal

if TYPE_CHECKING:
	import arc


def userify(user_id: int | str) -> str:
	"""### User mentions.

	```py
	>>> IFYs.userify(1234)
	'<@1234>'
	```
	"""

	return f"<@{user_id}>"


def userbangify(user_id: int | str) -> str:
	"""### User mentions (Old discord system).

	```py
	>>> IFYs.userbangify(1234)
	'<@!1234>'
	```
	"""

	return f"<@!{user_id}>"


def channelify(channel_id: int | str) -> str:
	"""### Channel mentions.

	```py
	>>> IFYs.channelify(1234)
	'<#1234>'
	```
	"""

	return f"<#{channel_id}>"


def roleify(role_id: int | str) -> str:
	"""### Role mentions.

	```py
	>>> IFYs.roleify(1234)
	'<@&1234>'
	```
	"""

	return f"<@&{role_id}>"


def commandify(command_name: str, command_id: int | str) -> str:
	"""### Command mentions.

	I'm not sure how to validate `command_name` other than the type (python regex apparently doesn't support whatever discord wrote [here](https://discord.com/developers/docs/interactions/application-commands#application-command-object-application-command-naming))

	```py
	>>> IFYs.commandify('settings', 1234)
	'</settings:1234>'
	```
	"""
	return f"</{command_name}:{command_id}>"


def emojify(emoji_name: str, emoji_id: int | str, *, animated: bool = False) -> str:
	"""### Emojis (animated and non-animated).

	I'm not sure how to validate `emoji_name` other than the type (i'm not aware of discord's emoji naming scheme unlime command_name where they at least published some regex? maybe they did somewhere but if they did as i said im not aware of it).

	```py
	>>> IFYs.emojify('mmLol', 1234)
	'<:mmLol:1234>'
	>>> IFYs.emojify('b1nzy', 1234, animated=True)
	'<a:b1nzy:1234>'
	```
	"""
	return f"<{'a' if animated else ''}:{emoji_name}:{emoji_id}>"


def timeify(timestamp: int, style: None | Literal["t", "T", "d", "D", "f", "F", "R"] = None) -> str:
	return f"<t:{timestamp}{':' if style is not None else ''}{style if style is not None else ''}>"


def specialify(option: Literal["customize", "browse", "guide"] = "customize") -> str:
	if option not in ("customize", "browse", "guide"):
		raise ValueError(f"Expected option in ('customize', 'browse', 'guide'), got {option} instead.")

	return f"<id:{option}>"


def ctx_commandify(ctx: "arc.Context") -> str:
	return f"<{ctx.command.display_name}:{ctx.interaction.command_id}>"



from collections.abc import Awaitable, Callable
from typing import Any, NoReturn


def void(*args: Any, **kwargs: Any) -> None:
	"""Synchronous voider: take any arguments and do nothing, useful in functions that require a callback when nothing is needed to be done."""


async def avoid(*args: Any, **kwargs: Any) -> None:
	"""Asynchronous voider: take any arguments and do nothing, useful in functions that require a callback when nothing is needed to be done."""


def raiser(e: Exception):
	"""With decorator-like structure return a synchronous callable which raises specified exception on call, no matter what (with *args, **kwargs which are ignored)."""

	def inner_raiser(*args, **kwargs) -> NoReturn:
		raise e

	return inner_raiser


async def araiser(e: Exception):
	"""With decorator-like structure return an asynchronous callable which raises specified exception on call, no matter what (with *args, **kwargs which are ignored)."""

	async def inner_raiser(*args, **kwargs) -> NoReturn:
		raise e

	return inner_raiser


def alambda[T, **P](f: Callable[P, T]) -> Callable[P, Awaitable[T]]:
	"""Asyncify any synchronous function, mostly lambdas when needed."""

	async def wrapper(*args, **kwargs):
		return f(*args, **kwargs)

	return wrapper

@lambda x: x()
class Void:
	def __call__(self, *args: Any, **kwargs: Any):
		return self

	def __getattr__(self, name):
		return self

	def __bool__(self):
		return False



"""# Useful stuff for tcr projects.

```
>>> from tcrutils.{feature_module} import feature
```

### Nyaaa :3

```
>>> from tcrutils.print import gay
>>> print(gay('nyaaaaa')) # uwu :3
```
"""

from ._version import __version__

__tcr_rainbow__ = True


if 0:  # A crude performance-measuring function
	import time

	def b(s=""):
		if not hasattr(b, "t"):
			b.t = time.perf_counter()
			b.s = 0
			return
		t1 = b.t
		t2 = time.perf_counter()

		t = t2 - t1
		b.s += t

		print(f"{s:10} {t:.10f}")
		b.t = time.perf_counter()


if False:  # Left `if False`d out for archival purposes (yeah.. there's git but i like it being here :3 uwu)
	from . import case as case
	from . import ensure_deps as ensure_depencencies
	from . import joke as joke
	from . import path, repl
	from . import types as types
	from .b64 import decode as b64d
	from .b64 import encode as b64e
	from .class_ import CachedInstancesMeta, DefaultsGetAttr, DefaultsGetItem, DefaultsGetSetAttr, DefaultsGetSetItem, NoInit, Singleton, new_cell, partial_class
	from .classfuncs import get_classname, get_name_classname, get_qualname_classname
	from .codeblock import codeblock, codeblocks, uncodeblock
	from .compare import able, able_simple, able_simple_result, isdunder
	from .console import breakpoint, console, start_eval_session
	from .console import console as c
	from .context import random_seed_lock
	from .decorator import copy_kwargs, copy_kwargs_sunder, skip_first_call, skip_first_call_async, test, timeit, with_overrides
	from .dev import generate_function_argument_typehints, generate_type_hinter
	from .dict import DotDict, clean_dunder_dict, dict_zip, merge_dicts
	from .dir import dir2, dir3, dir_recursive, vars2, vars3, vars_recursive
	from .extract_error import extract_error, extract_traceback, module_error_map, modules_error_map, print_exception_with_traceback
	from .getch import KeyCode, KeyCodeCompound, KeyCodeSimple, getch, getchs
	from .import_ import load_package_dynamically
	from .inject import ErrorCatcher, WarningCatcher
	from .input import insist
	from .inspect import eval_fback, get_file_colon_lineno, get_lineno
	from .int import clamp, float2int, hex, recursive_sum
	from .iterable import Or, batched, bogo_sort, cut_at, getattr_queue, getmanyattrs, hasmanyattrs, limited_iterable, shuffled, slice_between, stalin_sort
	from .language import apostrophe_s, make_plural, nth, plural_s
	from .misspellings import asert
	from .print import FMT_BRACKETS, FMTC, alert, fmt_iterable, gay, print_block, print_iterable
	from .regex import RegexPreset
	from .result import Result, ResultUnwrappedErrOnValueError, ResultUnwrappedOnErrorError
	from .run import RunSACAble, run_sac
	from .sdb import ShelveDB
	from .string import SlashableString, commafy, custom_zfill, get_token, join_urlstr, polaris_progressbar
	from .temp import temp_file
	from .terminal import terminal
	from .test_ import ass, raises, rass
	from .timestr import TStr, t_day, t_hour, t_minute, t_week, t_year, timestr
	from .typehints import force_keyword_only_typehints
	from .uptime import Uptime
	from .void import alambda, araiser, avoid, raiser, void



import os
from collections.abc import Generator as _Generator

if os.name == "nt":
	import msvcrt
else:
	import sys
	import termios
	import tty


class _KeyCodeBase:
	display_char_d: dict[str, str]

	def __init__(self):
		self.display_char_d = {}

	def __iter__(self):
		return (x for x in self.vars().values())

	def vars(self) -> dict[str, str]:
		return {k: v for k, v in self.__class__.__dict__.items() if k.isupper()}


class _KeyCodeSimpleType(_KeyCodeBase):
	if os.name == "nt":
		BACKSPACE = "\b"
		CTRL_BACKSPACE = "\b"  # Seems to be the same unfortunately? Cant differentiate between those :c

		WINDOWS_COMPOUND_KEYCODE_ESC_CHAR = "\xe0"
		"""(Probably) (In most cases) (i'm actually not sure) means: immediatelly following this press, there will be another, one indicating a compound keycode, for example the HOME key sends two pulses to getch, first b"\xe0" then immediately after b"G".

		This should be handled automatically by getchs()
		"""
	else:
		BACKSPACE = "\x7f"
		CTRL_BACKSPACE = "\x17"

	NULL = "\x00"

	ESC = "\x1b"

	CTRL_A = "\x01"
	CTRL_B = "\x02"
	CTRL_C = "\x03"
	CTRL_D = "\x04"
	CTRL_E = "\x05"
	CTRL_F = "\x06"
	CTRL_G = "\a"
	CTRL_H = "\b"
	CTRL_I = "\t"
	CTRL_J = "\n"
	CTRL_K = "\v"
	CTRL_L = "\f"
	CTRL_M = "\r"
	CTRL_N = "\x0e"
	CTRL_O = "\x0f"
	CTRL_P = "\x10"
	CTRL_Q = "\x11"
	CTRL_R = "\x12"
	CTRL_S = "\x13"
	CTRL_T = "\x14"
	CTRL_U = "\x15"
	CTRL_V = "\x16"
	CTRL_W = "\x17"
	CTRL_X = "\x18"
	CTRL_Y = "\x19"
	CTRL_Z = "\x1a"

	ENTER = CTRL_M
	TAB = CTRL_I

	def is_simple_ctrl(
		self,
		s: str,
		*,
		count_tab_and_enter_as_ctrl: bool = False,
		count_backspace_as_ctrl: bool = False,
	) -> bool:
		"""Is the given key, a key of the CTRL + any letter of the english alphabet (26 letters). Except Ctrl+I and Ctrl+M, since they map to the same as TAB and ENTER, though this can be adjusted via the keyword argument."""
		if count_tab_and_enter_as_ctrl:
			if s in (self.ENTER, self.TAB):
				return True

		if count_backspace_as_ctrl:
			if s == self.BACKSPACE:
				return True

		return s in (
			self.CTRL_A,
			self.CTRL_B,
			self.CTRL_C,
			self.CTRL_D,
			self.CTRL_E,
			self.CTRL_F,
			self.CTRL_G,
			# NO CTRL_H,
			# NO CTRL_I
			self.CTRL_J,
			self.CTRL_K,
			self.CTRL_L,
			# NO CTRL_M
			self.CTRL_N,
			self.CTRL_O,
			self.CTRL_P,
			self.CTRL_Q,
			self.CTRL_R,
			self.CTRL_S,
			self.CTRL_T,
			self.CTRL_U,
			self.CTRL_V,
			self.CTRL_W,
			self.CTRL_X,
			self.CTRL_Y,
			self.CTRL_Z,
		)


class _KeyCodeCompoundType(_KeyCodeBase):
	# Arrows
	if os.name == "nt":
		UP = "\0H"
		DOWN = "\0M"
		LEFT = "\0K"
		RIGHT = "\0P"
	else:
		UP = "\x1b[A"
		DOWN = "\x1b[B"
		LEFT = "\x1b[C"
		RIGHT = "\x1b[D"


class _KeyCodeType(_KeyCodeSimpleType, _KeyCodeCompoundType): ...


KeyCode = _KeyCodeType()
KeyCodeSimple = _KeyCodeSimpleType()
KeyCodeCompound = _KeyCodeCompoundType()


class _Getchs(_KeyCodeType):
	if os.name == "nt":

		def __call__(self):
			"""Reads a single character or key sequence from the keyboard."""
			first_char = msvcrt.getwch()

			if first_char in ("\x00", "\xe0"):
				second_char = msvcrt.getwch()
				return first_char + second_char
			return first_char

	else:

		def __call__(self):
			"""Reads a single character or key sequence from the keyboard."""
			fd = sys.stdin.fileno()
			old_settings = termios.tcgetattr(fd)
			try:
				tty.setraw(fd)
				char = sys.stdin.read(1)
				if char == "\x1b":
					return char + sys.stdin.read(2)
			finally:
				termios.tcsetattr(fd, termios.TCSADRAIN, old_settings)
			return char


getchs = _Getchs()
"""Improved implementation of tcr.getch(). Able to condense compound/[containg escape sequence] keypresses into one return value."""


class _Getch:
	"""Get a single character from standard input.

	Warning: May behave differently on Unix and Windows
	"""

	def __init__(self):
		try:
			self.impl = _GetchWindows()
		except ImportError:
			self.impl = _GetchUnix()

	def __call__(self):
		"""Get a single character from standard input.

		Warning: May behave differently on Unix and Windows
		"""
		return self.impl()


class _GetchUnix:
	def __init__(self):
		import sys
		import termios
		import tty

		self.sys = sys
		self.termios = termios
		self.tty = tty

	def __call__(self):
		fd = self.sys.stdin.fileno()
		old_settings = self.termios.tcgetattr(fd)
		try:
			self.tty.setraw(self.sys.stdin.fileno())
			ch = self.sys.stdin.read(1)
		finally:
			self.termios.tcsetattr(fd, self.termios.TCSADRAIN, old_settings)
		return ch.encode(encoding="ascii")


class _GetchWindows:
	def __init__(self):
		import msvcrt

		self.msvcrt = msvcrt

	def __call__(self):
		return self.msvcrt.getch()


getch = _Getch()
"""Original implementation of tcr.getch(). This is (probably) strictly worse than tcr.getchs(), please use that one instead, unless you really want to deal with escape syntax..."""



import re
from collections.abc import Callable, Iterable
from typing import Any, TypedDict

import attr as atr

from .console import console
from .run import run_sac


def _cut_string_at_indices(input_string, indices):
  return [input_string[i:j] for i, j in zip([0] + indices, indices + [None], strict=True)]


def _get_innermost_placeholders_indeces(text: str, R: str) -> list[int]:
  matches = re.finditer(R, text)
  spans = [match.span() for match in matches]
  return [item for t in spans for item in t]


def example_placeholder(name: str, *args: str, **kw) -> str:
  return name.upper()


def brackets_validator(inst, attr, value):
  if not isinstance(value, Iterable):
    raise TypeError('brackets must be an iterable')
  if not len(value) == 2:
    raise ValueError('brackets must be a tuple with two elements (opening and closing brackets)')
  if not all(isinstance(x, str) for x in value):
    raise TypeError('opening and closing brackets must be str')
  if len(value[0]) != len(value[1]):
    raise ValueError('opening and closing brackets must be the same length')


def placeholders_validator(inst, attr, value):
  if not isinstance(value, dict):
    raise TypeError('placeholders must be a dict')
  if not all(isinstance(x, str) for x in value):
    raise TypeError('placeholder keys must be str')
  if not all(callable(x) for x in value.values()):
    raise TypeError('placeholder keys must be str')


class MissingPlaceholderError(Exception): ...


@atr.s(kw_only=True)
class Execute:
  """r!execute from RoboTop."""

  brackets: tuple[str] = atr.ib(default=('{', '}'), validator=brackets_validator)
  splitter: str = atr.ib(default='|', validator=atr.validators.instance_of(str))
  placeholders: dict[str, Callable] = atr.ib(factory=dict, validator=placeholders_validator)

  def __attrs_post_init__(self) -> None:
    self.R = r'{[^{}]*}'.replace('{', re.escape(self.brackets[0])).replace('}', re.escape(self.brackets[1]))

  async def __call__(self, text, /, **kw):
    indices = _get_innermost_placeholders_indeces(text, self.R)

    if not indices:
      return text

    segments = _cut_string_at_indices(text, indices)

    segments = [((await self._execute_placeholder(x, **kw)) if (x.startswith(self.brackets[0]) and x.endswith(self.brackets[1])) else x) for x in segments]

    return await self(''.join(segments), **kw)

  async def _execute_placeholder(self, placeholder: str, **kw):
    splits = placeholder[len(self.brackets[0]) : -len(self.brackets[1])].split(self.splitter)
    name, *args = splits

    if name in self.placeholders:
      ret = await run_sac(self.placeholders[name], name, *args, **kw)
    else:
      ret = await self._default_placeholder(name, *args, **kw)

    return self._cleanse(ret)

  async def _default_placeholder(self, name, *ar, **kw):
    raise MissingPlaceholderError('Unknown placeholder: name')
    # return f'<unknown placeholder: {name!r}>'

  @staticmethod
  def _cleanse(text: Any) -> str:
    text = str(text)
    return text.replace('{', '[').replace('}', ']')

  def add_placeholder(self, name: str, func: Callable) -> None:
    placeholders_validator(None, None, {name: func})
    self.placeholders[name] = func



"""Contains joke functions or the ones that will never be used in a serious situations/prod (This entire package sucks so much it should never be used in prod but uhhhhh.....)."""

from typing import Literal, Self

from .print import FMT_BRACKETS
from .types import HexInt

fizzbuzz = lambda n: "Fizz" * (n % 3 == 0) + "Buzz" * (n % 5 == 0) or str(n)  # fmt: skip


def oddeven(n: int) -> Literal["odd", "even"]:
	"""### Outputs a string `'odd'` or `'even'` based on the supplied int `n`.

	```py
	>>> from tcrutils.other import oddeven
	>>> oddeven(1)
	'odd'
	>>> oddeven(2)
	'even'
	"""
	return "eovdedn"[n % 2 :: 2]


def christmas_tree(*, height=10, symbol="*"):
	"""Generate a christmas tree for printing in a console.

	Height is the number of lines the tree will have.\\
	Symbol is any 1-2 long string for example 'C#' or '*'.
	"""
	if len(symbol) == 1:
		symbol = 2 * symbol

	def transform(symbols: str):
		return symbols.center(height * 2, " ")

	return "\n".join([transform(symbol * x) for x in range(height + 1) if x])


def _cint_decorate_int_method(func):
	def wrapper(*args, **kwargs):
		result = func(*args, **kwargs)
		if isinstance(result, int):
			return CInt(result)
		else:
			return result

	return wrapper


class CInt:
	def __init__(self, value):
		self.value = value
		self.pos_flag = False
		self.neg_flag = False

	def __pos__(self):
		if self.pos_flag:
			self.value += 1
		self.pos_flag = not self.pos_flag
		return self

	def __neg__(self):
		if self.neg_flag:
			self.value -= 1
		self.neg_flag = not self.neg_flag
		return self

	def __eq__(self, other: Self | int) -> bool:
		if isinstance(other, CInt):
			return self.value == other.value
		else:
			return self.value == other

	def __int__(self):
		return self.value

	def __str__(self):
		return str(self.value)

	def __repr__(self) -> str:
		return self.__str__()

	def __tcr_fmt__(self, fmt_iterable, **kwargs):
		if self.neg_flag or self.pos_flag:
			return fmt_iterable(complex(int(self), 0.5 * (self.pos_flag - self.neg_flag)), **kwargs)
		else:
			return fmt_iterable(int(self), **kwargs)

	def __getattr__(self, name):
		if hasattr(self.value, name) and callable(getattr(self.value, name)):
			method = getattr(self.value, name)
			return _cint_decorate_int_method(method)
		else:
			raise AttributeError(f"'CInt' object has no attribute '{name}'")


class Private:
	"""Makes every attribute private, like in other languages that support private/public attribute modifiers.

	Unfortunately i haven't gotten to writing a Public version of it so you can only private every single field :3

	Usage:
	```py

	class PrivateString(Private, str): ...

	s = PrivateString('abc')

	print(s) # -> abc
	print(s.upper) # -> AttributeError: 'PrivateString' object has no attribute 'upper'
	print(s.__getattribute__) # -> AttributeError: 'PrivateString' object has no attribute '__getattribute__'

	```
	"""

	def __getattribute__(self, name):
		class_name = object.__getattribute__(self, "__class__").__name__
		raise AttributeError(f"{class_name!r} object has no attribute {name!r}")


@lambda x: x()
class gmail:
	def __init__(self) -> None:
		self.tlds = []

	def __getattr__(self, name):
		self.tlds.append(f"{name}")
		return self

	def __rmatmul__(self, other):
		class _Email:
			def __init__(self, email: str) -> None:
				self.email = email

			def __str__(self) -> str:
				return self.email

			def __repr__(self) -> str:
				return repr(self.__str__())

			def send(self, title: str, body: str) -> None:
				print("Sent email to " + self.email + " with title: " + repr(title) + ", and body: " + repr(body))

		try:
			return _Email(f"{other!s}@{self.__class__.__name__}.{'.'.join(self.tlds)}")
		finally:
			self.tlds.clear()


class Pointer:
	def __init__(self, value) -> None:
		self.__value = value

	def __iter__(self):
		yield self.__value

	def __next__(self):
		return self.__value

	def __repr__(self):
		value_type = type(self.__value).__name__

		address = hex(id(self.__value))

		return f"{self.__class__.__name__}<{value_type}>({address})"

	def __tcr_fmt__(self=None, *, fmt_iterable=None, no_addr=False, **kwargs):
		if self is None:
			raise NotImplementedError

		SH = kwargs.get("syntax_highlighting")

		this = lambda x, **kw: fmt_iterable(x, _raise_errors=True, **kwargs, **kw)

		addr = (FMT_BRACKETS[tuple][SH] % this(HexInt(id(self.__value)))) if not no_addr else ""

		if isinstance(self.__value, Pointer):
			type_ = self.__value.__tcr_fmt__(fmt_iterable=fmt_iterable, no_addr=True, **kwargs)
		else:
			type_ = this(self.__value.__class__)

		return f"{this(self.__class__)}{FMT_BRACKETS[range][SH] % type_}{addr}"


class __EchoType[T]:
	"""Echo the first argument of any interaction, like calling, gettingitem, gettingattr, etc.."""

	def __call__(self, x: T, /) -> T:
		return x

	def __getitem__(self, x: T, /) -> T:
		return x

	def __getattr__(self, x: T, /) -> T:
		return x

	def __getattribute__(self, x: T, /) -> T:
		return x

	def __add__(self, x: T, /) -> T:
		return x

	def __radd__(self, x: T, /) -> T:
		return x

	def __sub__(self, x: T, /) -> T:
		return x

	def __rsub__(self, x: T, /) -> T:
		return x

	def __mul__(self, x: T, /) -> T:
		return x

	def __rmul__(self, x: T, /) -> T:
		return x

	def __truediv__(self, x: T, /) -> T:
		return x

	def __rtruediv__(self, x: T, /) -> T:
		return x

	def __floordiv__(self, x: T, /) -> T:
		return x

	def __rfloordiv__(self, x: T, /) -> T:
		return x

	def __mod__(self, x: T, /) -> T:
		return x

	def __rmod__(self, x: T, /) -> T:
		return x

	def __matmul__(self, x: T, /) -> T:
		return x

	def __rmatmul__(self, x: T, /) -> T:
		return x

	def __divmod__(self, x: T, /) -> T:
		return x

	def __rdivmod__(self, x: T, /) -> T:
		return x

	def __pow__(self, x: T, /) -> T:
		return x

	def __rpow__(self, x: T, /) -> T:
		return x

	def __lshift__(self, x: T, /) -> T:
		return x

	def __rlshift__(self, x: T, /) -> T:
		return x

	def __rshift__(self, x: T, /) -> T:
		return x

	def __rrshift__(self, x: T, /) -> T:
		return x

	def __and__(self, x: T, /) -> T:
		return x

	def __rand__(self, x: T, /) -> T:
		return x

	def __xor__(self, x: T, /) -> T:
		return x

	def __rxor__(self, x: T, /) -> T:
		return x

	def __or__(self, x: T, /) -> T:
		return x

	def __ror__(self, x: T, /) -> T:
		return x

	def __get__(self, x, _owner=None):
		return x


echo = __EchoType()
del __EchoType



import datetime
import inspect
import re as regex
from collections.abc import Callable
from functools import partial, reduce
from sys import argv, exit

from colored import Back, Fore, Style

from .dict import clean_dunder_dict
from .extract_error import extract_error, extract_traceback
from .getch import getch
from .inspect import eval_fback, get_file_colon_lineno
from .iterable import cut_at
from .print import FMTC, fmt_iterable
from .terminal import terminal
from .types import QuotelessString


class CC:
	_ = Style.reset
	LOG = Fore.light_green + Style.bold  # 10
	WARN = Fore.YELLOW + Style.bold
	ERROR = Fore.RED + Style.bold
	DEBUG = Fore.purple_1b + Style.bold  # 129
	CRITICAL = Back.RED + Style.bold
	BANG = Fore.WHITE + Back.RED + Style.bold


class Console:
	"""### Provides logging capabilities.

	`console(...)` == `console.debug(...)`
	"""

	header_exprs: list[tuple[str, bool]]
	header_joiner: str
	header_end: str
	_last_diff: str | None = None
	include_callsite: bool = None

	def __init__(self, *, header_exprs: list[tuple[str, bool]] = None, header_joiner: str = ".", header_end: str = ": ") -> None:
		self.header_exprs = header_exprs or []
		self.header_joiner = header_joiner
		self.header_end = header_end

	@staticmethod
	def _get_timestamp():
		return str(datetime.datetime.now())[:-3].replace(".", ",")

	def _get_callsite_text_if_enabled(self, backtrack_frames) -> str:
		if self.include_callsite or (self.include_callsite is None and "--tcr-c-callsite" in argv):
			return get_file_colon_lineno(backtrack_frames=backtrack_frames)
		return ""

	def evaluate_header(self, fback_count: int = 5) -> str:
		result_list = [(expr if is_literal else eval_fback(expr, fback_count=fback_count)) for expr, is_literal in self.header_exprs]
		result_str = self.header_joiner.join(str(x) for x in result_list if x is not None)
		if result_str:
			result_str += self.header_end

		return result_str

	def _generate_out_and_print(self, *values, sep="\n", end="", withprefix=True, syntax_highlighting: bool = True, color: str, letter: str, _this_iteration_header: str | None = None, **kwargs) -> None:
		if _this_iteration_header is None:
			_this_iteration_header = self.evaluate_header()

		if not values:
			values = ("",)

		if len(values) > 1:
			for i, v in enumerate(values):
				if i == 0:
					char = "V"
				elif i == len(values) - 1:
					char = "Λ"
				else:
					char = "│"

				print(char, end=" ")

				self._generate_out_and_print(v, sep=sep, end=end, withprefix=withprefix, syntax_highlighting=syntax_highlighting, color=color, letter=letter, _this_iteration_header=_this_iteration_header, **kwargs)
			return

		values = [(f"{_this_iteration_header or ''}{x}" if isinstance(x, str) else fmt_iterable(x, syntax_highlighting=syntax_highlighting, **kwargs)) for x in values]

		out = sep.join(values)
		out = reduce(lambda x, y: str(x) + sep + str(y), [*values, ""])

		if withprefix:
			out = f"{letter} {(self._get_callsite_text_if_enabled(4) + ' ').lstrip()}{self._get_timestamp()} " + out

		out = f"{color if syntax_highlighting else ''}{out}{CC._ if syntax_highlighting else ''}"

		print(out, end=end)

	def log(self, *values, sep=" ", end="\n", withprefix=True) -> None | str:
		self._generate_out_and_print(*values, sep=sep, end=end, withprefix=withprefix, color=CC.LOG, letter="I")

	def warn(self, *values, sep=" ", end="\n", withprefix=True) -> None | str:
		self._generate_out_and_print(*values, sep=sep, end=end, withprefix=withprefix, color=CC.WARN, letter="W")

	def error(self, *values, sep=" ", end="\n", withprefix=True) -> None | str:
		self._generate_out_and_print(*values, sep=sep, end=end, withprefix=withprefix, color=CC.ERROR, letter="E")

	def critical(self, *values, sep=" ", end="\n", withprefix=True) -> None | str:
		self._generate_out_and_print(*values, sep=sep, end=end, withprefix=withprefix, color=CC.CRITICAL, letter="C")

	def debug(
		self,
		value: object = ...,
		/,
		*values: object,
		withprefix: bool = True,
		passthrough: bool = False,
		printhook: Callable[[str], None] = print,
		syntax_highlighting=True,
		margin: str = "",
		padding: str = " ",
		quoteless: bool = True,
		diff: bool = False,
		fmt_iterable: Callable[..., str] = fmt_iterable,
		eval: bool = False,
		**kwargs,
	) -> None | object:
		all_values = (value, *values)

		eval_header = self.evaluate_header(6)

		if eval_header:
			padding += eval_header

		if eval and len(all_values) == 1 and isinstance(all_values[0], str):
			pattern = regex.compile(r"^(.*?) ?(?:(?:=)|(?:->)|(?:=>)) ?$")
			match = pattern.match(all_values[0])

			if match:
				expression = match.group(1)

				result = eval_f_back(expression, f_backs=2)

				all_values = (all_values[0], result)
			else:
				raise ValueError("eval=True, but the only parameter's structure is invalid, must be a str that ends with '=', if you only want to pretty print, DO NOT USE THE eval=True KWARG!")
		elif eval:
			if len(all_values) != 1:
				raise ValueError("eval=True, but there is not exactly 1 positional parameter, if you only want to pretty print, DO NOT USE THE eval=True KWARG!")
			raise ValueError("eval=True, but the only parameter's structure is invalid, must be a str, if you only want to pretty print, DO NOT USE THE eval=True KWARG!")

		if len(all_values) >= 2 and all_values[0].__class__ is str and all_values[0]:
			first_string: str = all_values[0]
			after_first_string = ""

			found_any = False

			for symbol in ("=", "->", "=>"):
				if first_string.rstrip().endswith(symbol):
					found_any = True
					first_string = first_string.rstrip().removesuffix(symbol)
					if first_string and first_string[-1] == " ":
						after_first_string = f"{symbol} "
					else:
						after_first_string = symbol

			if found_any:
				if not withprefix and padding == " ":
					padding = ""
				padding += fmt_iterable(QuotelessString(first_string), syntax_highlighting=syntax_highlighting, **kwargs)
				padding += after_first_string

				quoteless = False
				all_values = all_values[1:]

		out = fmt_iterable(*[(x if ((not quoteless) or (x.__class__ is not str) or x == "" or x.strip() != x) else QuotelessString(x)) for x in all_values], syntax_highlighting=syntax_highlighting, **kwargs)

		if padding == " " and not withprefix:
			padding = ""

		prefix = ""
		if withprefix:
			prefix = f"D {(self._get_callsite_text_if_enabled(3) + ' ').lstrip()}{self._get_timestamp()}"

		c_debug = ""
		c_reset = ""
		c_bang = ""
		if syntax_highlighting:
			c_debug = CC.DEBUG
			c_reset = CC._
			c_bang = CC.BANG

		def compose(x) -> str:
			return f"{c_debug}{margin}{prefix}{padding}{x}{c_reset}"

		if diff and not self._last_diff:
			printhook(compose("\n" + "\n".join(["  " + x for x in out.split("\n")])))
		else:
			if not diff:
				printhook(compose(out))
			else:
				out_lines = out.split("\n")
				last_lines = self._last_diff.split("\n")

				while len(out_lines) < len(last_lines):
					last_lines.pop(-1)

				while len(last_lines) < len(out_lines):
					last_lines.append("")

				diff_out = "\n".join([("  " if line == line_old else f"{c_bang}*{c_reset} ") + line for line, line_old in zip(out_lines, last_lines, strict=True)])

				printhook(compose("\n" + diff_out))

		self._last_diff = out
		return None if not passthrough else value

	def hr(self, newlines_on_both_sides: bool = True, **kwargs):
		if newlines_on_both_sides:
			print()
		self.debug(QuotelessString("=" * terminal.width), withprefix=False, **kwargs)
		if newlines_on_both_sides:
			print()

	def with_expr_header(self, header_expr: str, literal: bool = False, joiner: str | None = None, end: str | None = None) -> "Console":
		if not isinstance(header_expr, str):
			raise TypeError("header_expr must be a string")

		if joiner is not None:
			if not isinstance(joiner, str):
				raise TypeError("overwrite_header_joiner must be a string")

		if end is not None:
			if not isinstance(end, str):
				raise TypeError("overwrite_header_end must be a string")

		return Console(
			header_exprs=self.header_exprs + [(header_expr, literal)],
			header_joiner=joiner if joiner is not None else self.header_joiner,
			header_end=end if end is not None else self.header_end,
		)

	def with_header(self, header: str):
		"""Add a simple header, alias for `with_expr_header(header, literal=True)`."""
		return self.with_expr_header(header, literal=True)

	__call__ = debug

	def __or__(self, other):
		return self.debug(other, passthrough=True)

	def __ror__(self, other):
		return self.debug(other, passthrough=True)

	def __repr__(self):
		return f"{self.__class__.__name__}()"

	def __lshift__(self, other):
		self.debug(other)
		return self

	def __rrshift__(self, other):
		self.debug(other)
		return self

	def __xor__(self, other):
		self.debug(other, withprefix=False)
		return other

	def __rxor__(self, other):
		self.debug(other, withprefix=False)
		return other


c = console = Console()
"""### Provides logging capabilities.

  `console(...)` == `console.debug(...)`
"""


def _clean_built_in_methods(d: dict) -> dict:
	return {k: v for k, v in d.items() if not isinstance(v, type(abs))}


def eval_f_back(code: str, *, f_backs: int = 1):
	"""Evaluates `code` in the context of the caller `f_backs` frames back."""

	frame = inspect.currentframe()
	for _ in range(f_backs):
		frame = frame.f_back

	return eval(code, frame.f_globals, frame.f_locals)


def start_eval_session(f_backs: int = 1) -> None:
	"""Starts an interactive shell. For debugging purposes."""
	try:
		current_frame = inspect.currentframe()
		for _ in range(f_backs):
			current_frame = current_frame.f_back
		desired_locals = current_frame.f_locals
		desired_globals = current_frame.f_globals

		a = " "
		while a:
			a = input("\n\r>>> ")
			if a:
				try:
					if a.strip().lower() == "locals":
						console(_clean_built_in_methods(clean_dunder_dict(desired_locals)))
					elif a.strip().lower() == "globals":
						console(_clean_built_in_methods(clean_dunder_dict(desired_globals)))
					elif a.strip().lower() == "code":
						lines, _ = inspect.findsource(current_frame)
						lineno = current_frame.f_lineno

						AROUND = 2
						CUTOFF = 70

						lines = lines[max(0, lineno - (AROUND + 1)) : lineno + AROUND]
						while all(x.startswith(" ") for x in lines):
							lines = [x[1:] for x in lines]
						lines = [f"{FMTC.NUMBER}{lineno - AROUND + i}{Fore.yellow + Style.bold} {'>' if i == AROUND else '|'}{FMTC._} {cut_at(x, CUTOFF)}" for i, x in enumerate(lines)]
						lines = "".join(lines)

						print(lines)
					else:
						console(eval(a, desired_locals, desired_globals))
				except Exception as e:
					console.error(f"\n\n{extract_traceback(e)}\n{extract_error(e)}")
	except (KeyboardInterrupt, EOFError):
		print("\r" + (" " * terminal.width) + "\r", end="")
		exit()


def breakpoint(*vals, printhook=console, clear=True, ctrlc: Callable[[], None] = start_eval_session) -> None:
	"""Stop the program execution until a key is pressed. Optionally pass in things to print."""
	vals and printhook(*vals)
	print(
		(a := f"{Fore.WHITE + Style.BOLD} >>> {Back.RED}BREAKPOINT{Style.RESET} {Fore.WHITE + Style.BOLD}<<<{Style.RESET}"),
		end="\r",
	)

	if getch() == b"\x03":
		if clear:
			print(len(a) * " ", end="")
		ctrlc()
		return

	if clear:
		print(len(a) * " ", end="\r")
	else:
		print()



import string

import colored

from ..console import console as c
from ..getch import getchs as getchs
from ..terminal import terminal
from .nodes import *
from .parser import parse_and_submit_nodes


def raise_for_not_only_nodes(nodes: tuple[Node]):
	if nodes and not all(isinstance(node, Node) for node in nodes):
		raise ValueError("nodes must be tuple[Node]")


class Repl:
	def display_node(self, n: Node, /) -> str:
		match n:
			case OwnDisplayNode():
				return n.display()
			case UnknownNode(text=text):
				return f"{colored.Style.UNDERLINE + colored.Fore.RED}{text}{FMTC._}"
			case CompoundNode():
				return "".join(self.display_node(x) for x in n._parsed)
			case Node(text=text):
				return f"{text}"
			case _:
				raise TypeError("Cannot display anything other than a Node")

	def __init__(self, *nodes: Node):
		raise_for_not_only_nodes(nodes)

		self.nodes = nodes
		self.chs_history = []
		self.chs_history_ptr = 0

	def on_ctrl(self, key: str) -> str | None:
		if key == getchs.CTRL_U:
			self.chs.clear()
			return
		if key == getchs.CTRL_C:
			raise KeyboardInterrupt

		if self.chs:
			return

		if key == getchs.CTRL_D:
			raise EOFError

	def on_arrow(self, key: str) -> str | None:
		if key == getchs.UP:
			if self.chs_history_ptr < self.chs_history.__len__() - 1:
				self.chs_history_ptr += 1
		if key == getchs.DOWN:
			if self.chs_history_ptr > 0:
				self.chs_history_ptr -= 1

	def printhook_prompt(self, last_char: str | None, *submitted_nodes: Node) -> tuple[bool, str]:
		fucked = bool(submitted_nodes and isinstance(submitted_nodes[-1], UnknownNode)) or contains_incomplete_nodes(submitted_nodes)

		chr1 = ">"

		if fucked:
			chr1 = "!"
		elif not submitted_nodes:
			chr1 = " "
		elif submitted_nodes[-1].children:
			chr1 = "+"

		chr2 = " " if chr1 != ">" else ">"

		if self.chs_history_ptr:
			chr3 = (str(self.chs_history_ptr)[-3:].__len__() - 1) * "\b" + str(self.chs_history_ptr)[-3:]

		chr3 = ">"
		return f"{chr1}{chr2}{chr3}"

	def printhook(self, last_char: str | None, *submitted_nodes: Node) -> None:
		print(f"{(terminal.width * ' ')}\r{self.printhook_prompt(last_char, *submitted_nodes)} {''.join(x.__str__() for x in submitted_nodes)}", end="\r")

	hide_cursor: bool = True
	valid_chars = string.ascii_letters + string.punctuation + string.digits + " \t" + "ąłćżźśćó"
	no_enter_on_unknown_or_incomplete: bool = True
	remove_root_node_from_output: bool = True

	@property
	def chs(self):
		return self.chs_history[self.chs_history_ptr]

	def __call__(self) -> list[Node]:
		while [] in self.chs_history:
			self.chs_history.remove([])

		self.chs_history.insert(0, [])
		self.chs_history_ptr = 0

		if self.hide_cursor:
			terminal.cursor.hide()

		self.printhook(None)

		try:
			while True:
				ch = getchs()

				if ch == getchs.ENTER:
					pass
				elif ch == getchs.BACKSPACE:
					if self.chs:
						self.chs.pop()
				elif ch == getchs.CTRL_BACKSPACE:
					if self.chs:
						while self.chs and self.chs[-1] in " \t":
							self.chs.pop()
						while self.chs and self.chs[-1] not in " \t":
							self.chs.pop()
						while self.chs and self.chs[-1] in " \t":
							self.chs.pop()
				elif ch in (getchs.UP, getchs.DOWN, getchs.LEFT, getchs.RIGHT):
					ch = self.on_arrow(ch)
				elif getchs.is_simple_ctrl(ch):
					ch = self.on_ctrl(ch)
				elif ch in self.valid_chars:
					self.chs.append(ch)

				submitted_nodes = parse_and_submit_nodes("".join(self.chs), self.nodes)

				self.printhook(ch, *submitted_nodes)

				if ch == getchs.ENTER:
					if self.no_enter_on_unknown_or_incomplete:
						if isinstance(submitted_nodes[-1], UnknownNode):
							continue

						if contains_incomplete_nodes(submitted_nodes):
							continue

					if submitted_nodes[-1].children:
						continue

					submitted_nodes = [x for x in submitted_nodes if not isinstance(x, DisposableNode)]
					submitted_nodes = [(x.convert() if isinstance(x, ConvertableNode) else x) for x in submitted_nodes]
					return submitted_nodes  # noqa: RET504
		finally:
			if self.hide_cursor:
				terminal.cursor.unhide()



import pytest
from tcrutils.result import Result, ResultUnwrappedErrOnValueError, ResultUnwrappedOnErrorError


def f(lhs: int, rhs: int) -> Result[float, ZeroDivisionError]:
	try:
		return Result.new_ok(lhs / rhs)
	except ZeroDivisionError as e:
		return Result.new_err(e)


@pytest.fixture
def good():
	return f(1, 2)


@pytest.fixture
def bad():
	return f(1, 0)


def test_good_is_ok(good):
	assert good.is_ok


def test_bad_is_err(bad):
	assert bad.is_err


def test_good_raises(good):
	with pytest.raises(ResultUnwrappedErrOnValueError):
		good.unwrap_err()


def test_bad_raises(bad):
	with pytest.raises(ResultUnwrappedOnErrorError):
		bad.unwrap()


def test_good_unwrap(good):
	assert isinstance(good.unwrap(), float)


def test_bad_unwrap(bad):
	assert isinstance(bad.unwrap_err(), ZeroDivisionError)


def test_unwrap_or(good, bad):
	X = object()
	assert good.unwrap_or(X) is not X
	assert bad.unwrap_or(X) is X


def test_unwrap_or_else(good, bad):
	X = object()
	assert good.unwrap_or_else(lambda _: X) is not X
	assert bad.unwrap_or_else(lambda _: X) is X


def test_unwrap_err_or(good, bad):
	X = object()
	assert good.unwrap_err_or(X) is X
	assert bad.unwrap_err_or(X) is not X


def test_unwrap_err_or_else(good, bad):
	X = object()
	assert good.unwrap_err_or_else(lambda _: X) is X
	assert bad.unwrap_err_or_else(lambda _: X) is not X



import abc
import datetime
import enum
import fnmatch
import os
import pathlib
import re
import sys
from copy import copy
from typing import Self

from ..console import console as c
from ..print import FMTC
from ..timestr import timestr_lookup
from .parser import parse_and_submit_nodes

PRINT_FORMAT_NODE_HTML_LIKE_OUTPUT = False


def contains_incomplete_nodes(submitted_nodes: tuple["Node"]):
	return any(isinstance(x, IncompleteNode) for x in submitted_nodes) or any(contains_incomplete_nodes(node._parsed or ()) for node in submitted_nodes if isinstance(node, CompoundNode))


class Node:
	__match_args__ = ("name", "text")

	name: str
	text: str | None

	children: list["Node | str"]
	parent: "Node | None"

	def __str__(self):
		return self.text

	def __tcr_fmt__(self=None, *, fmt_iterable, syntax_highlighting, **kwargs):
		if self is None:
			raise NotImplementedError

		if self.text is not None:
			body = fmt_iterable(()).replace(")", "") + fmt_iterable(self.text if not isinstance(self, ParsableNode) else self.parse()) + fmt_iterable(()).replace("(", "")
		else:
			at = f"{FMTC.DECIMAL}@" if syntax_highlighting else "@"
			body = (
				at
				+ fmt_iterable(self.name)
				+ fmt_iterable(
					self.children,
					let_no_indent=all(isinstance(x, str) for x in self.children),
					_force_tuple_no_trailing_comma_on_single_element=True,
				)
			)

		if PRINT_FORMAT_NODE_HTML_LIKE_OUTPUT:
			color = FMTC._
			op_bracket = f"{color}<"
			cl_bracket = f"{color}>"
			slash = f"{color}/"
			return f"{op_bracket}{FMTC.TYPE}{self.__class__.__name__}{cl_bracket}{body}{op_bracket}{slash}{FMTC.TYPE}{self.__class__.__name__}{cl_bracket}{FMTC._}"
		else:
			return f"{FMTC.TYPE}{self.__class__.__name__}{body}{FMTC._}"

	def __init__(self, *children: "Node", name: str = ""):
		if self.__class__ is Node:
			raise RuntimeError("Cannot instantiate Node without subclassing.")

		if not isinstance(name, str):
			raise TypeError(f"name must me a str, got {name.__class__!r}")

		self.name = name
		self.text = None
		self.parent = None

		self.children = list(children)
		for child in self.children:
			if isinstance(child, str):
				continue
			child.parent = self

	@abc.abstractmethod
	def match(self, s: str) -> tuple[str, str] | None:
		raise NotImplementedError("Must implement method match() for Node")

	def submit(self, s: str) -> None:
		self.text = s

	def copy_and_submit(self, s: str) -> Self:
		n = self.copy()
		n.submit(s)
		return n

	def copy(self) -> Self:
		return copy(self)

	def get_root_parent(self) -> "Node":
		current = self
		while (tmp := current.parent) is not None:
			current = tmp

		return current

	def resolve_path(self, path: str, *, _seen: None | set = None):
		if _seen is None:
			_seen = set()

		if path.startswith("/"):
			root = self.get_root_parent()
			nodes = [root]
		else:
			nodes = [self]

		parts = [x for x in path.split("/") if x]
		# if x removes consecutive and leading/trailing slashes

		for part in parts:
			if part == ".":
				continue
			if part == "..":
				nodes = [node.parent for node in nodes if node.parent is not None]
			else:
				matched = []
				for node in nodes:
					for child in node.children:
						if isinstance(child, Node):
							if fnmatch.fnmatch(child.name, part):
								matched.append(child)
						elif isinstance(child, str):
							if child in _seen:
								continue
							_seen.add(child)
							resolved_nodes = node.resolve_path(child, _seen=_seen)
							matched.extend(n for n in resolved_nodes if fnmatch.fnmatch(n.name, part))
				nodes = matched

		return nodes


### Bases


class ParsableNode[T](Node):
	__match_args__ = ("name", "value")

	def _init__(self, *args, **kwargs):
		if self.__class__ is ParsableNode:
			raise RuntimeError("Cannot instantiate Node without subclassing.")

		super().__init__(*args, **kwargs)

	exec("__init__ = _init__; del _init__")

	@abc.abstractmethod
	def parse(self) -> T:
		raise NotImplementedError("Must implement method parse() for ParsableNode")

	@property
	def value(self) -> T:
		return self.parse()


class DisposableNode(Node):
	"""Marked to be skipped when using the data, but not when displaying, for example you dont care about the mandatory space between keywords and identifiers but you still want to display it to the user."""


class ConvertableNode[T: Node](Node):
	"""Marked to be converted into another node with the convert method, after being displayed, before being returned to data processing."""

	@abc.abstractmethod
	def convert(self) -> T:
		raise NotImplementedError("Must implement method convert for ConvertableNode")


class OwnDisplayNode(Node):
	def display(self) -> str:
		return self.text


if True:  # Bases

	class _RegexNodeBase(Node):
		pattern: str | re.Pattern
		m: re.Match

		def __init_subclass__(cls, *, pattern: str | re.Pattern, **kwargs):
			cls.pattern = pattern
			super().__init_subclass__(**kwargs)

		def match(self, s):
			m = re.match(self.pattern, s)

			self.m = m

			if m is not None:
				return m.groups()[0], m.groups()[-1]
			return m

	class _StartswithNodeBase(Node):
		startswith_texts: list[str]

		def __init_subclass__(cls, *, startswith: tuple[str] | str, **kw):
			cls.startswith_texts = sorted((startswith,) if isinstance(startswith, str) else startswith, key=len, reverse=True)
			super().__init_subclass__(**kw)

		def match(self, s):
			for startswith_text in self.startswith_texts:
				if s.startswith(startswith_text):
					return (s[: len(startswith_text)], s[len(startswith_text) :])

			return None


class LiteralNode(Node):
	def __init__(self, name: str, *children: Node):
		super().__init__(*children, name=name)

	def match(self, s):
		if not s.startswith(self.name):
			return None

		return s[: len(self.name)], s[len(self.name) :]


class DisposableLiteralNode(DisposableNode, LiteralNode): ...


class UnreachableNode(Node):
	def match(self, s):
		return None


class IrrefutableNode(DisposableNode, Node):
	def match(self, s):
		return ("", s)


class MatchEverythingNode(Node):
	def __init__(self, **kwargs):
		super().__init__(**kwargs)

	def match(self, s):
		return s, ""


class KeywordNode(OwnDisplayNode):
	matching_text_escaped: str

	def __init__(self, name: str, *children: Node):
		super().__init__(*children, name=name)
		self.matching_text_escaped = re.escape(name)

	def match(self, s: str) -> tuple[str, str] | None:
		m = re.match(rf"^({self.matching_text_escaped})((?:\s|$).*)$", s)

		if m is not None:
			return m.groups()
		return m

	def display(self):
		return f"{FMTC.GD_COLON}{super().display()}{FMTC._}"


class AliasKeywordNode(KeywordNode, ConvertableNode[KeywordNode]):
	target_path: str

	def __init__(self, *children: Node, alias_name: str, target_path: str, autoadd_link_child: bool = True):
		super().__init__(alias_name, *children, *((f"{target_path}/*",) if autoadd_link_child else ()))
		self.target_path = target_path

	def convert(self):
		nodelist = self.resolve_path(self.target_path)

		if len(nodelist) != 1:
			if not nodelist:
				raise RuntimeError("AliasKeywordNode.convert(): was not pointing to a node (invalid path)")
			raise RuntimeError("AliasKeywordNode.convert(): was pointing to multiple nodes (valid path, but cannot point to >1 node)")

		if not isinstance(nodelist[0], KeywordNode):
			raise RuntimeError("AliasKeywordNode.convert(): was pointing to a valid node, but it's not a KeywordNode")  # noqa: TRY004

		return nodelist[0].copy_and_submit(self.text)


class UnknownNode(Node):
	children: list  # Cannot have children nodes

	def __init__(self, *, unknown_text: str):
		super().__init__(name=unknown_text)
		self.submit(unknown_text)
		self.children = []

	def match(self, s) -> tuple[str, str]:
		m = re.match(r"^(.*)()$", s)

		if m is not None:
			return m.groups()
		return m


class IncompleteNode(UnknownNode): ...


class EllipsisNode(OwnDisplayNode, _StartswithNodeBase, ParsableNode[Ellipsis], startswith="..."):
	def parse(self):
		return ...

	def display(self):
		return f"{FMTC.DECIMAL}{super().display()}{FMTC._}"


class IntNode(OwnDisplayNode, _RegexNodeBase, ParsableNode[int], pattern=r"^(-?\d+)(.*)$"):
	min: int | None = None
	max: int | None = None

	def __init_subclass__(cls, *, min: int = None, max: int = None):
		cls.min = min
		cls.max = max

	def match(self, s):
		result = super().m(s)
		if result is None:
			return None

		if self.min is None and self.max is None:
			return result

		parsed = int(result[0])

		if self.min is not None and self.min > parsed:
			return None

		if self.max is not None and self.max < parsed:
			return None

		return result

	def display(self):
		return f"{FMTC.NUMBER}{self.text}{FMTC._}"

	def parse(self):
		return int(self.text)


if True:  # Int-based Time & Date

	class Hour24IntNode(IntNode, min=0, max=23): ...

	class MinuteOrSecondIntNode(IntNode, min=0, max=59): ...

	class YearNode(IntNode, min=1000, max=9999): ...

	class MonthNode(IntNode, min=1, max=12): ...

	class DayNode(IntNode, min=1, max=31): ...


class FloatNode(OwnDisplayNode, _RegexNodeBase, ParsableNode[float], pattern=r"^(-?(?:\d+\.?\d*|\.?\d+))(.*)$"):
	def display(self):
		return f"{FMTC.NUMBER}{super().display()}{FMTC._}".replace(".", f"{FMTC.DECIMAL}.{FMTC.NUMBER}")

	def parse(self):
		return float(self.text)


class WordNode(OwnDisplayNode, _RegexNodeBase, pattern=r"^([a-zA-Z]+)(.*)$"): ...


class WordishNode(OwnDisplayNode, _RegexNodeBase, pattern=r"^([^ ]+)(.*)$"):
	def __init_subclass__(cls, **kwargs):
		return super().__init_subclass__(pattern=cls.pattern, **kwargs)


class PyIdentifierNode(_RegexNodeBase, pattern=r"^([a-zA-Z_][a-zA-Z0-9_]*)(.*)$"):
	def display(self):
		return f"{FMTC.DECIMAL}{super().display()}{FMTC._}"


class DoublequoteStrNode(OwnDisplayNode, _RegexNodeBase, pattern=r'^"((?:[^"\\]|\\.)*)"(.*)$'):
	def display(self):
		return f'{FMTC.QUOTES}"{FMTC.STRING}{super().display()}{FMTC.QUOTES}"{FMTC._}'


class SinglequoteStrNode(OwnDisplayNode, _RegexNodeBase, pattern=r"^'((?:[^'\\]|\\.)*)'(.*)$"):
	def display(self):
		return f"{FMTC.QUOTES}'{FMTC.STRING}{super().display()}{FMTC.QUOTES}'{FMTC._}"


class WordBreakNode(DisposableNode, _RegexNodeBase, pattern=r"^(\s|$)(.*)$"): ...


if True:  # Compound Node Basis

	class CompoundNodeMatchedMarkerNode(IrrefutableNode): ...

	class CompoundNode(Node):
		_parsed: None | list[Node]
		nodes: list[Node]
		require_matched_marker: bool

		def _init__(self, *args, **kwargs):
			if self.__class__ is CompoundNode:
				raise RuntimeError("Cannot instantiate CompoundNode without subclassing.")

			self._parsed = None

			super().__init__(*args, **kwargs)

		exec("__init__ = _init__; del _init__")

		def __tcr_fmt__(self=None, *, fmt_iterable, syntax_highlighting, **kwargs):
			if self is None:
				raise NotImplementedError

			if self.text is None:
				at = f"{FMTC.DECIMAL}@" if syntax_highlighting else "@"
				body = (
					fmt_iterable(self.nodes, _force_next_type=list)
					+ at
					+ fmt_iterable(self.name)
					+ fmt_iterable(
						self.children,
						let_no_indent=all(isinstance(x, str) for x in self.children),
						_force_tuple_no_trailing_comma_on_single_element=True,
					)
				)
			else:
				if isinstance(self, ParsableNode):
					try:
						parsed = self.parse()
					except Exception:
						parsed = "< FAILED TO PARSE >"

						if "--tcr-raise-errors" in sys.argv:
							raise
					body = fmt_iterable(()).replace(")", "") + fmt_iterable(parsed) + fmt_iterable(()).replace("(", "") + fmt_iterable([None] if self._parsed is None else [x for x in self._parsed if not isinstance(x, DisposableNode)])
				else:
					body = fmt_iterable(self._parsed)

			return f"{FMTC.TYPE}{self.__class__.__name__}{body}"

		def __init_subclass__(cls, nodes: list[Node], require_matched_marker: bool = False, **kwargs):
			cls.nodes = list(nodes)
			cls.require_matched_marker = require_matched_marker

			super().__init_subclass__(**kwargs)

		def match(self, s):
			parsed = parse_and_submit_nodes(s, self.nodes)

			if self.require_matched_marker and not any(isinstance(x, CompoundNodeMatchedMarkerNode) for x in parsed):
				return None

			match parsed[-1]:
				case UnknownNode():
					parsed = parsed[:-1]  # Accounted for with s.removeprefix(text), since text wouldnt contain the unknown node text.

			incomplete = bool(parsed and parsed[-1].children)

			self._parsed = list(parsed)

			text = "".join(x.text for x in parsed if not isinstance(x, IncompleteNode))

			if not text:
				return None

			return text, s.removeprefix(text), incomplete


if True:  # String

	class _StringDQuoteNode(DisposableNode, OwnDisplayNode, _RegexNodeBase, pattern=r'^(")(.*)$'): ...

	class _StringSQuoteNode(DisposableNode, OwnDisplayNode, _RegexNodeBase, pattern=r"^(')(.*)$"): ...

	class _StringDQContent(_RegexNodeBase, OwnDisplayNode, pattern=r'^((?:[^"\\]|\\.)*)(.*)$'): ...

	class _StringSQContent(_RegexNodeBase, OwnDisplayNode, pattern=r"^((?:[^'\\]|\\.)*)(.*)$"): ...

	class StringNode(
		OwnDisplayNode,
		CompoundNode,
		ParsableNode[str],
		nodes=(
			IrrefutableNode(
				UnreachableNode(
					_StringDQContent(
						_StringDQuoteNode(),
					),
					name="juncdq",
				),
				UnreachableNode(
					_StringSQContent(
						_StringSQuoteNode(),
					),
					name="juncsq",
				),
				_StringDQuoteNode(
					CompoundNodeMatchedMarkerNode("/juncdq/*"),
				),
				_StringSQuoteNode(
					CompoundNodeMatchedMarkerNode("/juncsq/*"),
				),
			),
		),
	):
		def __init_subclass__(cls, **kwargs):
			return super().__init_subclass__(cls.nodes, cls.require_matched_marker, **kwargs)

		def parse(self) -> str:
			parsed = [x for x in self._parsed if not isinstance(x, DisposableNode)]

			if len(parsed) != 1:
				raise RuntimeError("Malformed str parsing??? Expected only 1 Node left: _StringDQContent or _StringSQContent")

			return parsed[0].text.encode("utf-8").decode("unicode_escape")

		def display(self):
			s = ""
			for n in self._parsed:
				match n:
					case _StringDQuoteNode(text=text) | _StringSQuoteNode(text=text):
						s += f"{FMTC.QUOTES}{text}"
					case _StringDQContent(text=text) | _StringSQContent(text=text):
						s += f"{FMTC.STRING}{text}"
			return f"{s}{FMTC._}"


if True:  # List

	class _ListOpeningBracketNode(OwnDisplayNode, DisposableNode, _StartswithNodeBase, startswith="["):
		def display(self):
			return f"{FMTC.DECIMAL}{self.text}"

	class _ListClosingBracketNode(OwnDisplayNode, DisposableNode, _StartswithNodeBase, startswith="]"):
		def display(self):
			return f"{FMTC.DECIMAL}{self.text}{FMTC._}"

	class _ListCommaNode(OwnDisplayNode, DisposableNode, _StartswithNodeBase, startswith=","):
		def display(self):
			return f"{FMTC.DECIMAL}{self.text}{FMTC._}"

	class _GenericListNode[T](
		CompoundNode,
		ParsableNode[list[T]],
		require_matched_marker=True,
		nodes=(
			IrrefutableNode(
				UnreachableNode(
					WordBreakNode("../*"),
					_ListCommaNode("/content/*"),
					_ListClosingBracketNode(),
					name="junc",
				),
				IrrefutableNode(
					CompoundNodeMatchedMarkerNode(
						_ListOpeningBracketNode(
							"/content/*",
							WordBreakNode("../*"),
							_ListClosingBracketNode(),
						),
					),
				),
			),
		),
	):
		_symlink_to_content_node: UnreachableNode

		def _init__(self, *args, **kwargs):
			if self.__class__ is _GenericListNode:
				raise RuntimeError("Cannot instantiate _GenericListNode without subclassing.")

			return super().__init__(*args, **kwargs)

		exec("__init__ = _init__; del _init__")

		def parse(self) -> list:
			parsed = [x for x in self._parsed if not isinstance(x, DisposableNode)]

			l = []

			for node in parsed:
				l.append(node.parse())

			return l

		def __init_subclass__(cls, listof):
			listof = [x(f"listitem{i}", "/junc/*") for i, x in enumerate(listof)]

			content_unreachable_node = UnreachableNode(
				*listof,
				WordBreakNode(
					"../*",
				),
				_ListClosingBracketNode(),
				name="content",
			)
			cls._symlink_to_content_node = content_unreachable_node

			content_unreachable_node.parent = cls.nodes[0]
			cls.nodes[0].children.append(content_unreachable_node)

	class ListOfInt(_GenericListNode[int], listof=(IntNode,)): ...

	class ListOfStr(_GenericListNode[str], listof=(StringNode,)): ...

	class ListOfStrOrInt(_GenericListNode[int | str], listof=(IntNode, StringNode)): ...


if True:  # Timestr

	class TimestrNodeRobostrPartNode(
		CompoundNode,
		ParsableNode[float],
		require_matched_marker=True,
		nodes=(
			FloatNode(
				CompoundNodeMatchedMarkerNode(
					*[LiteralNode(unit, "*", IrrefutableNode()) for unit in sorted(timestr_lookup, key=len, reverse=True)],
				),
			),
		),
	):
		def parse(self):
			seconds = 0

			parsed = [x for x in self._parsed if not isinstance(x, DisposableNode)]

			for node_number_int, node_unit_literal in zip(parsed[::2], parsed[1::2], strict=True):
				match (node_number_int, node_unit_literal):
					case (ParsableNode(_, n), LiteralNode(u)):
						seconds += n * timestr_lookup[u]
					case _:
						raise RuntimeError("Invalid set of nodes")

			return seconds

	class TimestrNodeTIMEPartNode(
		CompoundNode,
		ParsableNode[tuple[int, int, int]],
		require_matched_marker=True,
		nodes=(
			Hour24IntNode(
				DisposableLiteralNode(
					":",
					CompoundNodeMatchedMarkerNode(
						MinuteOrSecondIntNode(
							DisposableLiteralNode(
								":",
								MinuteOrSecondIntNode(),
								IrrefutableNode(),
							),
							IrrefutableNode(),
						),
						IrrefutableNode(),
					),
				),
			),
		),
	):
		tzinfo: datetime.tzinfo

		def _init__(self, *args, **kwargs):
			super().__init__(*args, **kwargs)
			self.with_tzinfo()

		exec("__init__ = _init__; del _init__")

		def with_tzinfo(self, tzinfo: datetime.tzinfo = datetime.datetime.now().astimezone().tzinfo):
			self.tzinfo = tzinfo
			return self

		def parse(self):
			parsed = [x.parse() for x in self._parsed if not isinstance(x, DisposableNode)]

			h, m, s, *_ = (*parsed, 0, 0)  # Fill unspecified values with zeroes, eg. 13: -> 13:00:00

			return (h, m, s)

	class TimestrNodeDATEPartNode(
		CompoundNode,
		ParsableNode[tuple[int, int, int]],
		require_matched_marker=True,
		nodes=(
			DayNode(
				DisposableLiteralNode(
					"-",
					CompoundNodeMatchedMarkerNode(
						MonthNode(
							DisposableLiteralNode(
								"-",
								YearNode(),
								IrrefutableNode(),
							),
							IrrefutableNode(),
						),
						IrrefutableNode(),
					),
				),
			),
		),
	):
		tzinfo: datetime.tzinfo

		def _init__(self, *args, **kwargs):
			super().__init__(*args, **kwargs)
			self.with_tzinfo()

		exec("__init__ = _init__; del _init__")

		def with_tzinfo(self, tzinfo: datetime.tzinfo = datetime.datetime.now().astimezone().tzinfo):
			self.tzinfo = tzinfo
			return self

		def _try_parse(self, d: str, m: str, y: str):
			if not m or not y:
				now = datetime.datetime.now(tz=self.tzinfo)

			if not m:
				m = now.month

			if not y:
				y = now.year

			y, m, d = int(y), int(m), int(d)

			return (y, m, d)

		def match(self, s):
			result = super().match(s)

			if result is None:
				return None

			matched, *_ = result

			d, m, y, *_ = (*matched.split("-"), "", "")

			if self._try_parse(d, m, y) is None:
				return None

			return result

		def parse(self):
			parsed = [x.parse() for x in self._parsed if not isinstance(x, DisposableNode)]

			d, m, y, *_ = (*parsed, "", "")

			return self._try_parse(d, m, y)

	class TimestrNode(
		CompoundNode,
		ParsableNode[float],
		nodes=(
			UnreachableNode(
				KeywordNode(
					"!",
					"/*",
				),
				name="junc",
			),
			TimestrNodeTIMEPartNode(
				CompoundNodeMatchedMarkerNode("/junc/*", IrrefutableNode()),
			),
			TimestrNodeDATEPartNode(
				CompoundNodeMatchedMarkerNode("/junc/*", IrrefutableNode()),
			),
			TimestrNodeRobostrPartNode(
				CompoundNodeMatchedMarkerNode("/junc/*", IrrefutableNode()),
			),
		),
	):
		def _init__(self, *args, **kwargs):
			super().__init__(*args, **kwargs)
			self.with_tzinfo()

		exec("__init__ = _init__; del _init__")

		def with_tzinfo(self, tzinfo: datetime.tzinfo = datetime.datetime.now().astimezone().tzinfo):
			for node in self.children:
				if hasattr(node, "with_tzinfo"):
					node.with_tzinfo(tzinfo)
			return self

		def match(self, s):
			result = super().match(s)

			if len(tuple(x for x in self._parsed if isinstance(x, TimestrNodeTIMEPartNode))) > 1:
				return None

			return result

		def parse(self):
			parsed = (x for x in self._parsed if not isinstance(x, DisposableNode))

			seconds = 0

			node_time_result = None
			node_date_result = None

			for node in parsed:
				match node:
					case TimestrNodeRobostrPartNode(_, value):
						seconds += value
					case TimestrNodeTIMEPartNode(_, value):
						# Guaranteed to ever be only 1 or 0 of them, due to TimestrNode.match()
						node_time_result = value
					case TimestrNodeDATEPartNode(_, value):
						# Guaranteed to ever be only 1 or 0 of them, due to TimestrNode.match()
						node_date_result = value
					case n:
						raise RuntimeError(f"[BUG] Unknown node: {n!r}")

			print("tcrutils:nodes.py:TimestrNode | warning, timestr isnt finished yet dont use it...")

			# now = datetime.datetime.now(tz=tzinfo)

			# then = datetime.datetime.combine(node_date_result, node_time_result)

			# seconds += (then - now).total_seconds()

			return seconds


# if True:  # TStrNode

# 	class _TStrDateNode(_RegexNodeBase, ParsableNode[tuple[int, int, int]], pattern=r"^((?P<d>\d?\d)-(?:(?P<m>\d?\d)(?:-(?P<y>(?:\d\d)?\d\d)?)?)?)(.*)$"):
# 		def match(self, s):
# 			result = super().match(s)

# 			return result

# 		@staticmethod
# 		def _validate_by_str_parts(d: str, m: str | None, y: str | None):
# 			try:
# 				d, m, y = int(d), int(m), int(y)
# 				datetime.date(y, m, d)
# 			except Exception:
# 				return False
# 			else:
# 				return True

# 		def parse(self):
# 			groups = self.m.groupdict()

# 			d, m, y = groups["d"], groups["m"], d["y"]

# 	class TStrNode(
# 		CompoundNode,
# 		ParsableNode[float],
# 		require_matched_marker=True,
# 		nodes=(
# 			IrrefutableNode(
# 				UnreachableNode(
# 					LiteralNode("!", "/*"),
# 					name="separator",
# 				),
# 				UnreachableNode(
# 					name="content",
# 				),
# 			),
# 		),
# 	):
# 		tzinfo: datetime.tzinfo

# 		def __init__(self, *children, tzinfo: datetime.tzinfo = datetime.datetime.now().astimezone().tzinfo, name: str = ""):
# 			super().__init__(*children, name=name)
# 			self.tzinfo = tzinfo


if True:  # Path Stuff

	class _PathWordishNode(WordishNode, OwnDisplayNode):
		def display(self):
			text = self.text.replace("/", f"{FMTC.PATH_SLASH}/{FMTC.STRING}")

			if os.name == "nt":
				text = text.replace("\\", f"{FMTC.PATH_SLASH}\\{FMTC.STRING}")

			return f"{FMTC.STRING}{text}"

	class _PathStringNode(StringNode):
		def display(self):
			text = super().display().replace("/", f"{FMTC.PATH_SLASH}/{FMTC.STRING}")

			if os.name == "nt":
				text = text.replace("\\", f"{FMTC.PATH_SLASH}\\{FMTC.STRING}")

			return f"{text}"

	class PurePathNode(
		CompoundNode,
		ParsableNode[pathlib.PurePath],
		require_matched_marker=True,
		nodes=(
			IrrefutableNode(
				_PathStringNode(CompoundNodeMatchedMarkerNode()),
				_PathWordishNode(CompoundNodeMatchedMarkerNode()),
			),
		),
	):
		def __init_subclass__(cls, **kwargs):
			return super().__init_subclass__(cls.nodes, cls.require_matched_marker, **kwargs)

		def parse(self):
			parsed = [x for x in self._parsed if not isinstance(x, DisposableNode)]

			match parsed:
				case [WordishNode(_, s) | StringNode(_, s)]:
					return pathlib.PurePath(s)
				case _:
					raise RuntimeError(f"Invalid matched sequence?: {parsed}")

	class PathNode(PurePathNode, ParsableNode[pathlib.Path]):
		class ValidState(enum.Flag):
			EXISTS_FILE = 1 << 0
			EXISTS_DIRECTORY = 1 << 1
			DOESNT_EXIST = 1 << 2

			@classmethod
			def all(cls):
				return cls.DOESNT_EXIST | cls.EXISTS_DIRECTORY | cls.EXISTS_FILE

		valid_state: ValidState
		filename_regex: str | re.Pattern

		def __init__(
			self,
			*children,
			valid_state: ValidState = ValidState.all(),
			filename_regex: str | re.Pattern = r"^.*$",
			**kwargs,
		):
			super().__init__(*children, **kwargs)
			self.valid_state = valid_state
			self.filename_regex = filename_regex

		def _validate_filename(self, path: pathlib.Path):
			return bool(re.match(self.filename_regex, path.name))

		def _validate_state(self, path: pathlib.Path):
			if not path.exists() and (self.valid_state & self.ValidState.DOESNT_EXIST):
				return True

			if path.is_file() and (self.valid_state & self.ValidState.EXISTS_FILE):
				return True

			if path.is_dir() and (self.valid_state & self.ValidState.EXISTS_DIRECTORY):  # noqa: SIM103
				return True

			return False

		def match(self, s):
			m = super().match(s)

			if contains_incomplete_nodes(self._parsed):
				return m

			path = self.parse()

			if not all((
				self._validate_filename(path),
				self._validate_state(path),
			)):
				self._parsed.insert(0, IncompleteNode(unknown_text="!{"))
				self._parsed.append(IncompleteNode(unknown_text="}"))

			return m

		def parse(self):
			return pathlib.Path(super().parse()).resolve()
